// use this in class to help students identify the FOCUS
//   (item to count), see countBPmatches++ below;
//
// then find the recurrence relation from the recursion



int isPalDNAwithERROR(char* f, char* e, int m)
{
	/* Recursive algorithm to determine if a string is an INVERTED REPEAT (IR) */
	/* WITH upto 'm' mismatches, finds DNA definition of a palindrome (IR), e.g., "AC*T*AGT" */
	
	/* mismatches are allowed in basepairing; on a mismatch, three other tries are made:
		(1) move in left-end,  leave right-end as is, add +1 to number of mismatches so far
		(2) move in right-end, leave left-end as is,  add +1 to number of mismatches so far
		(3) move in both ends, add +2 to number of mismatches so far
	
	/* tripley recursive algorithm :(  */
	/* recurrence relation, F, for the number of times we call basePairsMatch() 
	   as a function of length, L:  
	
					0 	 						0<=L<=1 
			F(L) = 	2*F(L-1) + F(L-2) + 1		L>1   	
			
	   ARGS:
	   			f	- pointer to left-most position in the string
	   			e	- pointer to right-most position in the string
	   			m	- number of mismatches so far
	   			
	   GLOBALS:	MaxError - maximum number of mismatches allowed
	   
	   RETURNS: true if the original f points to an IR with MaxError mismatches or less,
	   			false otherwise
	   			
	   			true is returned on the first successful IR found; this does not find the
	   			optimal IR (where optimal might be defined as the fewest number of mismatches)
	*/
	
	if (m > MaxError)
		return FALSE;					/* no IR; too many mismatches */
			
	else if (f >= e)
			return TRUE;				/* voila, we have an IR */
		
	else 
	{
		countBPmatches++;
		if ( basePairsMatch(*f, *e) )					/* check these two base pairs == */
		{
			if ( isPalDNAwithERROR(f+1, e-1, m) )	/* move both ends in if they match */
				return TRUE;
			else
				return FALSE;	
		}
	
		else											/* otherwise */
		{
		
			if ( isPalDNAwithERROR(f+1, e, m+1) )		/* move in left-end */
				return TRUE;
			
			else if ( isPalDNAwithERROR(f, e-1, m+1) )	/* move in right-end */
				return TRUE;
			
			else if ( isPalDNAwithERROR(f+1, e-1, m+2) ) /* move in both ends */
				return TRUE;
	
			else
				return FALSE;							 /* nope, no IR found */
		}
	}
	
}
