#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "constants.h"
#include "prototypes_IRR.h"

/* uncomment the next line out if you only want near-EXCEL ready output
   for the number of =='s performed in the isPalDNAwithError() function */
//#define WORSECASE_EXCEL


extern int location;

extern long MaxError;

extern double countBPmatches;


void findPalsInBuffer(char* Buffer, int Bsize, int length)
{
   /* int junk; */
   int i,j;
   char palString[MAX_LEN+1];
   
   double after, before;
   
/*
    printf("---------------------------------\n");
	puts(Buffer);
    printf("---------------------------------\n");
*/
   
   

/*
   printf("*** start looking for Pal's at location = %d\n", *location); 
*/

   /*========== try worse case runs thru the i loop ===================================*/
   #ifdef WORSECASE_EXCEL
   for (i=2; i <= 20; i++)
   {
   		for (j=0; j<i; j++)
   			palString[j] = 'A';   /* fill with WORSE CASE all A's */
		palString[i] = '\0';
		
		MaxError = i-2;  /* simulate % error for each string */
		
		/* careful: we can't have 100% error since if f==e then we consider that success
		the largest absolute number of mismatches that we can handle is length-2 mismatches 
		if (MaxError >= i-2)
		  MaxError = i-2;
		*/
		  
		countBPmatches = 0;
   		isPalDNAwithERROR(palString, palString+i-1, 0);
		
		printf("%20s\t%03d\t%e\n", palString, i, countBPmatches);
		
	}
	
	#else  /*===== regular run working thru file =======================================*/

    for (i=0; i < (Bsize - (length-1)); i++)
    {
   		memcpy(palString, &Buffer[i], length);
   		palString[length] = '\0';

        printf("*** searching string %s at location %d\n", palString, i+(location));  

   		/* if (isPalDNA(palString))    (used this before we handled error) */
   		
   		before = countBPmatches;
   		//if (isPalDNAwithERROR(palString, palString+length-1, 0))
		if (isPalDNAwithERROR_printSTARs(palString, palString+length-1, 0))
			printPal(palString, i);
		after = countBPmatches;
		/* printf("\n++++++++ This IR check took %e (==) checks\n", after-before); */
		
   }
   location = location + i;
   
   #endif   /*===========================================================================*/

/*
  printf("\t\t**** NEXT location to check is location = %d\n", *location);
  
  printf("\n-------------- end of Buffer ---------------\n\n"); 
*/
}

int isPalDNA(char *f)
{
	/* finds DNA definition of a palindrome, e.g., "acgtacgt" */
	char bpair;
	char* e;
	
	e = f + (strlen(f) - 1);
	while (f < e)
	{
		switch(*f)
		{
			case 'A':  bpair = 'T'; break;
			case 'C':  bpair = 'G'; break;
			case 'G':  bpair = 'C'; break;
			case 'T':  bpair = 'A'; break;
			default:   printf("\n\n*** ERROR:  invalid nucleotide char %c found \n", *f); 
		}
		if (*e == bpair)
		 {
		 	f++;
		 	e--;
		 }
		else
			return FALSE;
	}
	return TRUE;
}



int isPalDNAwithERROR(char* f, char* e, int m)
{
	/* Recursive algorithm to determine if a string is an INVERTED REPEAT (IR) */
	/* WITH upto 'm' mismatches, finds DNA definition of a palindrome (IR), e.g., "AC*T*AGT" */
	
	/* mismatches are allowed in basepairing; on a mismatch, three other tries are made:
		(1) move in left-end,  leave right-end as is, add +1 to number of mismatches so far
		(2) move in right-end, leave left-end as is,  add +1 to number of mismatches so far
		(3) move in both ends, add +2 to number of mismatches so far
	
	/* tripley recursive algorithm :(  */
	/* recurrence relation, F, for the number of times we call basePairsMatch() 
	   as a function of length, L:  
	
					0 	 						0<=L<=1 
			F(L) = 	2*F(L-1) + F(L-2) + 1		L>1   	
			
	   ARGS:
	   			f	- pointer to left-most position in the string
	   			e	- pointer to right-most position in the string
	   			m	- number of mismatches so far
	   			
	   GLOBALS:	MaxError - maximum number of mismatches allowed
	   
	   RETURNS: true if the original f points to an IR with MaxError mismatches or less,
	   			false otherwise
	   			
	   			true is returned on the first successful IR found; this does not find the
	   			optimal IR (where optimal might be defined as the fewest number of mismatches)
	*/
	
	if (m > MaxError)
		return FALSE;					/* no IR; too many mismatches */
			
	else if (f >= e)
			return TRUE;				/* voila, we have an IR */
		
	else 
	{
		countBPmatches++;
		if ( basePairsMatch(*f, *e) )					/* check these two base pairs == */
		{
			if ( isPalDNAwithERROR(f+1, e-1, m) )	/* move both ends in if they match */
				return TRUE;
			else
				return FALSE;	
		}
	
		else											/* otherwise */
		{
		
			if ( isPalDNAwithERROR(f+1, e, m+1) )		/* move in left-end */
				return TRUE;
			
			else if ( isPalDNAwithERROR(f, e-1, m+1) )	/* move in right-end */
				return TRUE;
			
			else if ( isPalDNAwithERROR(f+1, e-1, m+2) ) /* move in both ends */
				return TRUE;
	
			else
				return FALSE;							 /* nope, no IR found */
		}
	}
	
}

int isPalDNAwithERROR_printSTARs(char* f, char* e, int m)
{
	/* this version keeps track of each ERROR location and replaces it with a '*'  */
	
	/* Recursive algorithm to determine if a string is an INVERTED REPEAT (IR) */
	/* WITH upto 'm' mismatches, finds DNA definition of a palindrome (IR), e.g., "AC*T*AGT" */
	
	/* mismatches are allowed in basepairing; on a mismatch, three other tries are made:
		(1) move in left-end,  leave right-end as is, add +1 to number of mismatches so far
		(2) move in right-end, leave left-end as is,  add +1 to number of mismatches so far
		(3) move in both ends, add +2 to number of mismatches so far
	
	/* tripley recursive algorithm :(  */
	/* recurrence relation, F, for the number of times we call basePairsMatch() 
	   as a function of length, L:  
	
					0 	 						0<=L<=1 
			F(L) = 	2*F(L-1) + F(L-2) + 1		L>1   	
			
	   ARGS:
	   			f	- pointer to left-most position in the string
	   			e	- pointer to right-most position in the string
	   			m	- number of mismatches so far
	   			
	   GLOBALS:	MaxError - maximum number of mismatches allowed
	   
	   RETURNS: true if the original f points to an IR with MaxError mismatches or less,
	   			false otherwise
	   			
	   			true is returned on the first successful IR found; this does not find the
	   			optimal IR (where optimal might be defined as the fewest number of mismatches)
	*/
	
	if (m > MaxError)
		return FALSE;					/* no IR; too many mismatches */
			
	else if (f >= e)
			return TRUE;				/* viola, we have an IR */
		
	else 
	{
		countBPmatches++;
		if ( basePairsMatch(*f, *e) )					/* check these two base pairs == */
		{
			if ( isPalDNAwithERROR_printSTARs(f+1, e-1, m) )	/* move both ends in if they match */
				return TRUE;
			else
				return FALSE;	
		}
	
		else											/* otherwise */
		{
			// change left to '*'
			char old = *f;
			*f = '*';
			if ( isPalDNAwithERROR_printSTARs(f+1, e, m+1) )		/* move in left-end */
			{
				return TRUE;
			}
			else 
			{
				//char old;
				*f = old;
				old = *e;
				*e = '*';
				if ( isPalDNAwithERROR_printSTARs(f, e-1, m+1) )	/* move in right-end */
				{
					return TRUE;
				}
				else 
				{
					*e = old;
					char old_f = *f;
					char old_e = *e;
					*f = '*';
					*e = '*';
					if ( isPalDNAwithERROR_printSTARs(f+1, e-1, m+2) ) /* move in both ends */
					{
						return TRUE;
					}
					else
					{
						*f = old_f;
						*e = old_e;
						return FALSE;							 /* nope, no IR found */
					}
				}
			}
		}
	}
	
}



int basePairsMatch(char a, char b)
{
	int match = FALSE;
	
	switch(a)
	{
			case 'A':  match = (b == 'T'); break;
			case 'C':  match = (b == 'G'); break;
			case 'G':  match = (b == 'C'); break;
			case 'T':  match = (b == 'A'); break;
			default:   /*printf("\n\n*** ERROR:  invalid nucleotide char %c found \n", a); 
                       system("PAUSE"); exit(-1); */ ;
	}
	
	return match;
}



void printPal(char palString[], int offset)
{
 /* location assumed GLOBAL */

   	int spot, line, charspot;

	spot = offset + (location);
	line = spot/MAX_LINE+1 +1;  /* extra +1 is to count header */
	charspot = spot%MAX_LINE;
   	printf("%cPalindrome %s at character #%d\n", 7, palString, spot);
   	printf("\tLine #%d, character #%d\n", line, charspot);
/*
   	scanf("%d", &junk);  	
*/
}

